"""
GSM Widgets Package

This package provides interactive widgets for visualizing and manipulating
GSM (Generalized Standard Material) models using ipywidgets and ipytree.
"""

from .gsm_def_widget import GSMDefWidget, create_gsm_widget

__all__ = ['GSMDefWidget', 'create_gsm_widget']
