"""
GSM Lagrange Framework

This package provides a symbolic-to-executable framework for thermodynamically
consistent material modeling based on the Generalized Standard Material concept.
"""

__version__ = '0.1.0'

# Simple imports without lazy loading to avoid issues
# These can be imported individually as needed:
# from .core import GSMDef, GSMEngine, GSMModel, Scalar, Vector, ResponseData
# from .models import GSM1D_ED, GSM1D_EP, etc.
# from .cli import cli_gsm, data_structures, etc.
